import tempfile
import os
import glob
import socket
import yaml
import logging

from fpdf import FPDF
from fpdf.enums import XPos, YPos
from PyPDF2 import PdfFileWriter, PdfFileReader, PdfFileMerger

from .XnatUtils import get_assessor_inputs
from .utilities import read_yaml
from .assessor_utils import parse_full_assessor_name, is_sgp_assessor
from .version import VERSION as dax_version

# TODO: recode this so the arguments are just the text to use
# for the header and footer parts, and we figure
# out the text outside of this function

# TODO: allow option to load description from the singularity README file.
# should this be the default if there's not one explicitly in the yaml?
# What would be the default location? This could default to use the
# "main container" if it is set.

LOGGER = logging.getLogger('dax')


def make_overpdf(overfile, info, pagenum, pagecount):
    # Assessor name top left of header
    # processor name and version left of footer
    # Date and Page numbers in right footer

    # Initialize the pdf
    pdf = FPDF(orientation="P", unit='in', format='letter')
    pdf.set_margins(left=0, top=0, right=0)
    pdf.add_page()
    pdf.set_font('courier', size=9)

    # Draw lines in header and footer
    pdf.set_draw_color(r=155, g=155, b=155)
    pdf.set_line_width(0.01)
    pdf.line(x1=0.2, y1=0.2, x2=8.3, y2=0.2)
    pdf.line(x1=0.2, y1=10.7, x2=8.3, y2=10.7)

    # Write the assessor label in the header
    pdf.set_xy(0.3, 0.1)
    pdf.cell(0, 0.05, '{}'.format(info['assessor']))

    # Prevent footer from falling off page
    pdf.set_auto_page_break(False)

    # Write proc version in left of footer
    pdf.set_xy(0.3, -0.2)
    pdf.cell(0, 0, '{}'.format(info['proctype']))

    # Write date in right of footer
    pdf.set_xy(-2.1, -0.2)
    pdf.cell(0, 0, '{}'.format(info['procdate']))

    # Write page numbers right of footer
    pdf.set_xy(-0.6, -0.2)
    pdf.cell(0, 0, '{}/{}'.format(pagenum, pagecount))

    # Write the pdf to file
    pdf.output(overfile)


def make_lastpdf(lastfile, info):
    pdf = FPDF(orientation="P", unit='in', format='letter')
    pdf.add_page()

    LOGGER.debug('making last page for PDF')

    # Session Info
    pdf.ln(0.05)
    for key, val in info['session'].items():
        pdf.set_font('helvetica', size=13)
        pdf.cell(w=1.1, h=.4, text=key, border=0)
        pdf.set_font('courier', size=14)
        pdf.cell(w=6.7, h=.4, text=val, border=1, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Inputs
    pdf.ln(0.5)
    pdf.set_font('helvetica', size=14)
    pdf.cell(1, .2, 'INPUTS:', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Inputs - data rows
    for r in info['inputs']:
        pdf.set_font('courier', size=12)
        pdf.cell(w=1.25, h=.3, text=r[0], border=0)
        pdf.cell(w=1, h=.3, text=r[1][-6:], border=1, align='C')
        pdf.set_font('courier', size=7)
        pdf.cell(w=6, h=.3, text=(r[2].split('/assessors/')[-1])[-90:], border=0, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Outputs
    pdf.ln(0.5)
    pdf.set_font('helvetica', size=14)
    pdf.cell(1, .2, 'OUTPUTS:', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Outputs - header row
    pdf.set_font('helvetica', style='B', size=12)
    pdf.cell(w=1.5, h=.3, text='resource', border=1)
    pdf.cell(w=0.6, h=.3, text='type', border=1, align='C')
    pdf.cell(w=5.6, h=.3, text='path', border=1, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Outputs - data rows
    pdf.set_font('courier', size=12)
    for out in info['outputs']:
        pdf.cell(w=1.5, h=.3, text=out['resource'], border=1)
        pdf.cell(w=0.6, h=.3, text=out['type'], border=1, align='C')
        pdf.cell(w=5.6, h=.3, text=out['path'], border=1, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Job Info
    pdf.ln(0.5)
    joby = pdf.get_y()
    pdf.set_font('helvetica', size=14)
    pdf.cell(1, .2, 'JOB:', new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    pdf.set_font('courier', size=12)
    for k, v in info['job'].items():
        # Show label
        pdf.set_font('helvetica', size=12)
        pdf.cell(w=0.8, h=.3, text=k, border=0)

        # Show value
        pdf.set_font('courier', size=12)
        pdf.cell(w=1, h=.3, text=v, border=1, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Proc Info
    pdf.set_y(joby)
    pdf.set_font('helvetica', size=14)
    pdf.set_x(2.5)
    pdf.cell(1, .2, '', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    for k, v in info['proc'].items():
        pdf.set_x(2.5)
        pdf.set_font('helvetica', size=12)
        pdf.cell(w=1.2, h=.3, text=k, border=0)
        pdf.set_font('courier', size=12)
        pdf.cell(w=4.4, h=.3, text=v, border=1, new_x=XPos.LMARGIN, new_y=YPos.NEXT)

    # Show Description
    pdf.ln(0.5)
    pdf.set_font('helvetica', size=14)
    pdf.cell(0, .1, 'DESCRIPTION:', new_x=XPos.LMARGIN, new_y=YPos.NEXT)
    pdf.set_font('courier', size=10)
    pdf.multi_cell(0, .2, text=info['description'])

    # Save to file
    LOGGER.debug('saving last page of PDF to file:{}'.format(lastfile))
    try:
        pdf.output(lastfile)
    except Exception as err:
        LOGGER.error('saving PDF:{}:{}'.format(lastfile, err))


def make_suppdf(outfile, info, infile=None, overlay=True):
    # Make the overlay PDF
    tempdir = tempfile.mkdtemp()
    lastfile = os.path.join(tempdir, 'last.pdf')
    mergedfile = os.path.join(tempdir, 'merged.pdf')

    # Make last pages of pdf
    try:
        make_lastpdf(lastfile, info)
    except Exception as err:
        import traceback
        traceback.print_stack()
        LOGGER.error('error making PDF:{}:{}'.format(lastfile, err))

    # Append last page(s)
    merger = PdfFileMerger()

    # Start with the input pdf if we have one
    if infile:
        merger.append(PdfFileReader(infile))

    # Append out last page(s)
    merger.append(PdfFileReader(lastfile))

    # Write the merged PDF file
    merger.write(mergedfile)

    # Load the merged PDF
    newpdf = PdfFileReader(open(mergedfile, "rb"))
    pagecount = newpdf.getNumPages()

    # Initialize the new pdf
    output = PdfFileWriter()

    # Iterate pages in existing PDF to add overlays
    for i in range(newpdf.getNumPages()):
        pagenum = i + 1
        overfile = os.path.join(tempdir, '{}.pdf'.format(pagenum))
        try:
            make_overpdf(overfile, info, pagenum, pagecount)
        except Exception as err:
            ('error making PDF', err)

        # Load the overlay PDF we just made
        overpdf = PdfFileReader(overfile)

        # Merge the two
        page = newpdf.getPage(i)
        page.mergePage(overpdf.getPage(0))

        # Append merged page to output
        output.addPage(page)

    # Write to file
    with open(outfile, "wb") as f:
        output.write(f)


def get_this_instance():
    # build the instance name
    this_host = socket.gethostname().split('.')[0]
    this_user = os.environ['USER']
    return '{}@{}'.format(this_user, this_host)


def load_version(assr_path):
    with open(os.path.join(assr_path, 'version.txt'), 'r') as f:
        return f.read().strip()


def load_attr(assr_path, attr):
    filepath = os.path.join(
            os.path.dirname(assr_path),
            'DISKQ',
            attr,
            os.path.basename(assr_path))

    with open(filepath, 'r') as f:
        return f.readline().strip()


def load_description(assr_path):
    processor_path = load_attr(assr_path, 'processor')
    with open(processor_path, "r") as yaml_stream:
        doc = yaml.load(yaml_stream, Loader=yaml.FullLoader)

    return doc.get('description', '')


def load_procyamlversion(assr_path):
    processor_path = load_attr(assr_path, 'processor')
    LOGGER.debug('loading procyamlversion from file:{}'.format(processor_path))
    with open(processor_path, "r") as yaml_stream:
        doc = yaml.load(yaml_stream, Loader=yaml.FullLoader)

    return doc.get('procyamlversion', '')


def load_inputs(assr_obj):
    assr_inputs = get_assessor_inputs(assr_obj)
    inputs = []

    for k, v in assr_inputs.items():
        inputs.append((k, v.split('/')[-1], v))

    return inputs


def load_outputs(assr_path):
    outputs = []
    processor_path = load_attr(assr_path, 'processor')

    doc = read_yaml(processor_path)

    for c in doc.get('outputs', []):
        LOGGER.debug('suppdf:load_outputs:{}'.format(c))
        # Check for keywords
        if 'pdf' in c:
            _path = c['pdf']
            _type = 'FILE'
            _res = 'PDF'
        elif 'dir' in c:
            _path = c['dir']
            _type = 'DIR'
            _res = c['dir']
        elif 'stats' in c:
            _path = c['stats']
            _type = 'FILE'
            _res = 'STATS'

        # Get explicitly set path, type, resource
        # These will override anything set by keywords
        if 'path' in c:
            _path = c['path']

        if 'type' in c:
            _type = c['type']

        if 'resource' in c:
            _res = c['resource']

        # Add to our outputs list
        outputs.append({
            'path': _path,
            'type': _type,
            'resource': _res})

    return outputs


def load_info(assr_path, assr_obj):
    info = {}

    # Load info
    assr = parse_full_assessor_name(os.path.basename(assr_path))

    LOGGER.debug('loading info:{}'.format(assr))

    info['assessor'] = assr['label']

    # Load
    LOGGER.debug('loading version')
    info['procdate'] = load_attr(assr_path, 'jobstartdate')

    # Load job info
    LOGGER.debug('loading job info')
    info['job'] = {
        'jobid': load_attr(assr_path, 'jobid'),
        'duration': load_attr(assr_path, 'walltimeused'),
        'memory': load_attr(assr_path, 'memused')}

    # Load proc info
    LOGGER.debug('loading proc info')
    info['proc'] = {
        'dax_version': str(dax_version),
        'dax_manager': get_this_instance()}

    # Load procyaml version
    LOGGER.debug('suppdf:load_info:load_procyamlversion')
    info['procyamlversion'] = load_procyamlversion(assr_path)

    # Load the description from the processor yaml
    LOGGER.debug('suppdf:load_info:load_description')
    info['description'] = load_description(assr_path)

    # Load inputs from XNAT assessor/inputs
    LOGGER.debug('suppdf:load_info:load_inputs')
    info['inputs'] = load_inputs(assr_obj)

    # Load outputs from the processor yaml
    LOGGER.debug('suppdf:load_info:load_outputs')
    info['outputs'] = load_outputs(assr_path)

    if is_sgp_assessor(os.path.basename(assr_path)):
        info['proctype'] = info['assessor'].split('-x-')[2]
        info['session'] = {
            'PROJECT': assr['project_id'],
            'SUBJECT': assr['subject_label']}
    else:
        info['proctype'] = info['assessor'].split('-x-')[3]
        info['session'] = {
            'PROJECT': assr['project_id'],
            'SUBJECT': assr['subject_label'],
            'SESSION': assr['session_label']}

    LOGGER.debug('suppdf:load_info:finished')
    return info


def suppdf(assr_path, assr_obj):
    try:
        info = load_info(assr_path, assr_obj)
        if info['procyamlversion'] != '3.0.0-dev.0':
            LOGGER.debug('skipping suppdf:{}, procyamlversion={}'.format(
                assr_path, info['procyamlversion']))
            return False
    except Exception as err:
        LOGGER.debug('skipping suppdf:{}:err={}'.format(assr_path, err))
        return False

    LOGGER.debug('suppdf:={}'.format(info))

    # find the input pdf
    try:
        inputpdf = glob.glob(assr_path + '/PDF/*.pdf')[0]
    except Exception:
        inputpdf = None

    # Make sure the output dir exists
    try:
        os.mkdir(os.path.join(assr_path, 'PDF'))
    except FileExistsError:
        pass

    # Name the output with assessor name
    outputpdf = '{}/PDF/report_{}.pdf'.format(assr_path, info['assessor'])

    if inputpdf == outputpdf:
        LOGGER.debug('PDF already supped:{}'.format(inputpdf))
        return

    # supplement it by adding header/footer overlays
    make_suppdf(outputpdf, info, inputpdf, overlay=True)

    # delete the input pdf
    if inputpdf and os.path.exists(outputpdf):
        LOGGER.debug('deleting input PDF:{}'.format(inputpdf))
        os.remove(inputpdf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import ArgumentParser, RawDescriptionHelpFormatter
from dax import XnatUtils
import ast
import os
import pandas as pd
import pyxnat
import re
import requests

host = os.environ['XNAT_HOST']
user = 'admin'


def parse_args():
    argp = ArgumentParser(prog='SwitchProjects', formatter_class=RawDescriptionHelpFormatter)
    argp.add_argument('--proj', dest='project', default=None, help='Project we want to push from')
    argp.add_argument('--txt', dest='txtFile', default=None, help='Name of file we want to store session list')
    return argp


if __name__ == '__main__':
    PARSER = parse_args()
    OPTIONS = PARSER.parse_args()

    project = OPTIONS.project
    f = open(OPTIONS.txtFile,'a')

    with XnatUtils.get_interface(host=host, user=user) as XNAT:
        subjects = XNAT.get_subjects(project)

        for subject in subjects:
            sessions = XNAT.get_sessions(project, subject['label'])
            for session in sessions:
                scans = XNAT.get_scans(project, subject['label'], session['label'])
                df = pd.DataFrame(scans)
                df = df[['subject_label','session_label','scan_type','scan_label']]
                if 'unknown' in str(df.loc[0]):
                    print('ExamCard found')
                else:
                    patient_id = session['session_label'].split('_')
                    request = 'http://10.109.20.19:8080/dcm4chee-arc/aets/DCM4CHEE57/rs/studies/?PatientID=*{}'.format(patient_id[1])
                    res = requests.get(request)
                    instances = ast.literal_eval(res.text)
                    for j, instance in enumerate(instances):
                        instanceUID = instance['00081190']['Value'][0]
                        instanceUID = instanceUID + '/series'
                        res = requests.get(instanceUID)
                        experiment = ast.literal_eval(res.text)
                        for k, instances in enumerate(experiment):
                            if '0008103E' not in instances:
                                url = instances['00081190']['Value'][0]
                                url = url + '/export/VandyXNAT'
                                res = requests.post(url)
                                print('*****************************')
                                print(patient_id[1])
                                f.write(session['subject_label'] + ' ' + session['session_label'] + ' ' + patient_id[1] + '\n')

    f.close()



# flake8: noqa
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from . import bin
from . import dax_tools_utils
from . import log
from . import xnat_tools_utils
from . import XnatUtils

from .task import Task
from .cluster import PBS
from .launcher import Launcher
from .dax_settings import DAX_Settings, DAX_Netrc
from .version import VERSION as __version__
from .git_revision import git_revision as __git_revision__
from .XnatUtils import SpiderProcessHandler, AssessorHandler
from .modules import ScanModule, SessionModule
from .processors import AutoProcessor



import dax.XnatUtils as XnatUtils

from unittest import TestCase

from dax.task import (JOB_FAILED, JOB_RUNNING, JOB_PENDING, READY_TO_UPLOAD,
                   NEEDS_QA, RERUN, REPROC, FAILED_NEEDS_REPROC, BAD_QA_STATUS)

from . import unit_test_common_processor_yamls as processor_yamls

bad_qa_status = [JOB_PENDING, NEEDS_QA, REPROC, RERUN, FAILED_NEEDS_REPROC]


class FakeXnat:

    @staticmethod
    def placeholder():
        raise NotImplementedError()

    @staticmethod
    def get_proctype(spider, suffix=None):
        # TODO: ok to use XnatUtils method, at least until it is removed from XnatUtils, which should happen
        return XnatUtils.get_proctype(spider, suffix)

    @staticmethod
    def is_cscan_unusable(scan):
        return scan.info()['quality'] == 'unusable'

    @staticmethod
    def is_cscan_usable(scan):
        return scan.info()['quality'] == 'usable'

    @staticmethod
    def has_resource(scan, label):
        return XnatUtils.has_resource(scan, label)

    @staticmethod
    def get_good_cassr(csess, sp_types, needs_qc):
        return XnatUtils.get_good_cassr(csess, sp_types, needs_qc)


class TestResource:

    def __init__(self):
        self.thing = None

# class TestAssessorObject:
#
#     def __init__(self, label):
#         self.label = label
#
#
#     def info(self):
#         return {'label': self.label}


class TestProjectObject:
    def __init__(self, project_dict):
        project_id = project_dict['ID']
        self.project_id_ = project_id
        self.label_ = project_dict['label']
        self.subject_objects_ = {
            subj['ID']: TestSubjectObject(self, subj)
            for subj in project_dict['subjects']
        }

    def subjects(self):
        return self.subject_objects_

    def project_id(self):
        return self.project_id_

    def label(self):
        return self.label_


class TestSubjectObject:
    def __init__(self, project, subject_dict):
        self.project_ = project
        #self.project_id_ = project_id
        subject_id = subject_dict['ID']
        self.subject_id_ = subject_id
        self.label_ = subject_dict['label']
        self.session_objects_ = {
            sess['ID']: TestSessionObject(self, sess)
            for sess in subject_dict['sessions']
        }

    def sessions(self):
        return self.session_objects_

    def project_id(self):
        return self.project_.project_id()

    def subject_id(self):
        return self.subject_id_

    def parent(self):
        return self.project_

    def label(self):
        return self.label_


class TestSessionObject:
    def __init__(self, subject, session_dict):
        self.subject_ = subject
        session_id = session_dict['ID']
        self.session_id_ = session_id
        self.label_ = session_dict['label']
        self.scan_objects_ = {
            scan['ID']: TestScanObject(self, scan)
            for scan in session_dict['scans']
        }
        self.assessor_objects_ = {
            assr['ID']: TestAssessorObject(self, assr)
            for assr in session_dict['assessors']
        }

    def project_id(self):
        return self.subject_.project_id()

    def subject_id(self):
        return self.subject_.subject_id()

    def session_id(self):
        return self.session_id_

    def label(self):
        return self.label_

    def session(self):
        return self

    def parent(self):
        return self.subject_

    def info(self):
        return {
            'project_id': self.subject_.project_id(),
            'subject_label': self.subject_.subject_id(),
            'session_label': self.session_id_
        }

    @staticmethod
    def entity_type():
        return 'session'

    def has_shared_project(self):
        return None

    def scan_by_key(self, key):
        return self.scan_objects_[key]

    def assessor_by_key(self, key):
        return self.assessor_objects_[key]

    def scans(self):
        return self.scan_objects_

    def assessors(self):
        return self.assessor_objects_


class TestScanObject:

    def __init__(self, session, scan):
        self.quality_ = scan['quality']
        self.session_ = session
        scan_id = scan['ID']
        self.scan_id_ = scan_id
        self.resource_objects_ = {
            r['label']: TestResourceObject(self, r)
            for r in scan['resources']
        }
        self.type_ = scan['type']

    def project_id(self):
        return self.session_.project_id()

    def subject_id(self):
        return self.session_.subject_id()

    def session_id(self):
        return self.session_.session_id()

    def scan_id(self):
        return self.scan_id_

    def label(self):
        return self.scan_id_

    def type(self):
        return self.type_

    def session(self):
        return self.session_

    def parent(self):
        return self.session_

    def info(self):
        return {
            'ID': self.scan_id_,
            'project_id': self.session_.project_id(),
            'subject_label': self.session_.subject_id(),
            'session_label': self.session_.session_id(),
            'scan_label': self.scan_id_,
            'quality': self.quality_
        }

    @staticmethod
    def entity_type():
        return 'scan'

    def resources(self):
        return list(self.resource_objects_.values())

    def __getitem__(self, key):
        return self.resource_objects_[key]


class TestAssessorObject:

    def __init__(self, session, assessor_dict):
        self.session_ = session
        assessor_id = assessor_dict['ID']
        self.assessor_id_ = assessor_id
        self.label_ = assessor_dict['label']
        self.resource_objects_ = {
            r['label']: TestResourceObject(assessor_id, r)
            for r in assessor_dict['resources']
        }
        self.proctypes_ = assessor_dict['proctypes']
        self.qcstatus_ = assessor_dict['qcstatus']
        self.procstatus_ = assessor_dict['procstatus']

    def project_id(self):
        return self.session_.project_id()

    def subject_id(self):
        return self.session_.subject_id()

    def session_id(self):
        return self.session_.session_id()

    def assessor_id(self):
        return self.assessor_id_

    def label(self):
        return self.label_

    def type(self):
        return self.proctypes_

    def session(self):
        return self.session_

    def parent(self):
        return self.session_

    def info(self):
        return {
            'project_id': self.session_.project_id(),
            'subject_label': self.session_.subject_id(),
            'session_label': self.session_.session_id(),
            'label': self.label_,
            'proctype': self.proctypes_,
            'qcstatus': self.qcstatus_,
            'procstatus': self.procstatus_
        }

    def usable(self):
        return not self.qcstatus_ in bad_qa_status

    def unusable(self):
        return self.qcstatus_ in bad_qa_status

    def resources(self):
        return list(self.resource_objects_.values())

    @staticmethod
    def entity_type():
        return 'assessor'

    def __getitem__(self, key):
        return self.resource_objects_[key]


class TestResourceObject:
    def __init__(self, scan, resource_dict):
        self.scan_ = scan
        self.resource_dict_ = resource_dict

    def project_id(self):
        return self.scan_.project_id()

    def subject_id(self):
        return self.scan_.subject_id()

    def session_id(self):
        return self.scan_.session_id()

    def scan_id(self):
        return self.scan_.scan_id()

    def label(self):
        return self.resource_dict_['label']

    def __getitem__(self, key):
        return self.resource_dict_[key]


class TestObjectsTest(TestCase):

    def test_make_repo(self):
        # t = TestSessionObject(
        #     'proj1', 'subj1',
        #     brain_tiv_from_gif_xnat_contents['projects'][0]['subjects'][0]['sessions'][0])
        tp = TestProjectObject(brain_tiv_from_gif_xnat_contents['projects'][0])
        print((tp.label()))
        for tsb in tp.subjects().values():
            print(('', tsb.label()))
            for tse in tsb.sessions().values():
                print((' ', tse.label()))
                for tsc in tse.scans().values():
                    print(('  ', tsc.label()))
                for tas in tse.assessors().values():
                    print(('  ', tas.label()))


# second attempt; cut at xnatutils, assessorhandler and interfacetemp

# TODO: refactor calls to InterfaceTemp.select: move to methods on InterfaceTemp
# TODO: create FakeAssessorHandler
# TODO: create FakeXnatInterface


# class processor_yamls:
#     scan_gif_parcellation = processor_yamls.scan_gif_parcellation_yaml
#     scan_brain_tiv_from_gif = processor_yamls.scan_brain_tiv_from_gif_yaml


spider_tiv_from_gif = '/home/dax/Xnat-management/comic100_dax_config/pipelines/BrainTivFromGIF/v1.0.0/Spider_BrainTivFromGIF_v1_0_0.py'

brain_tiv_from_gif_xnat_contents = {
    'projects': [{
        'ID': 'proj1',
        'label': 'project 1',
        'subjects': [{
            'ID': 'subj1',
            'label': 'subject 1',
            'sessions': [{
                'ID': 'sess1',
                'label': 'session 1',
                'scans': [{
                    'ID': '1',
                    'type': 'T1',
                    'quality': 'usable',
                    'resources': [{
                        'label': 'SNAPSHOTS',
                        'file_count': 2
                    }, {
                        'label': 'NIFTI',
                        'file_count': 1
                    }]
                }, {
                    'ID': '2',
                    'type': 'T1',
                    'quality': 'usable',
                    'resources': [{
                        'label': 'SNAPSHOTS',
                        'file_count': 1
                    }, {
                        'label': 'NIFTI',
                        'file_count': 1
                    }]
                }],
                'assessors': [{
                    'ID': 'proc1',
                    'label': 'proj1-x-subj1-x-sess1-x-1-x-proc1',
                    'proctypes': 'GIF_Parcellation_v3',
                    'procstatus': 'COMPLETE',
                    'qcstatus': 'Passed QA',
                    # 'resources': [{
                    #     'label': 'PDF',
                    #     'file_count': 1
                    # }, {
                    #     'label': 'pBRAIN',
                    #     'file_count': 1
                    # }, {
                    #     'label': 'bBRAIN',
                    #     'file_count': 1
                    # }, {
                    #     'label': 'pTIV',
                    #     'file_count': 1
                    # }, {
                    #     'label': 'bTIV',
                    #     'file_count': 1
                    # }, {
                    #     'label': 'SNAPSHOTS',
                    #     'file_count': 2
                    # }, {
                    #     'label': 'PBS',
                    #     'file_count': 1
                    'resources': [{
                        'label': 'SEG',
                        'file_count': 1
                    }]
                }]
            }]
        }]
    }]
}

xnat_contents = {
    'brain_tiv_from_gif': brain_tiv_from_gif_xnat_contents
}

git_pct_t1_yaml = """
---
inputs:
  default:
    spider_path: /home/dax/Xnat-management/comic100_dax_config/pipelines/GIF_pCT/v1.0.0/Spider_GIF_pCT_v1_0_0.py
    working_dir: /scratch0/dax/
    nipype_exe: perform_gif_propagation.py
    db: /share/apps/cmic/GIF/1946_database_pCT/db_t1.xml
  xnat:
    scans:
      - name: scan1
        types: T1w,MPRAGE,T1,T1W,1946_3DT1
        resources:
          - resource: NIFTI
            varname: t1
      - name: scan2
        types: 1946_MRAC_UTE
        resources:
          - resource: NIFTI
            varname: ute_echo2
      - name: scan3
        types: 1946_MRAC_UTE_UMAP
        resources:
          - resource: DICOM
            varname: ute_umap
      - name: scan4
        types: RECON_PRR_NAC Images
        resources:
          - resource: DICOM
            varname: pet
command: python {spider_path} --t1_file {t1} --ute_echo2_file {ute_echo2} --ute_umap_dcm {ute_umap} --nac_pet_dcm {pet} --dbt {db} --working_dir {working_dir} --exe {nipype_exe}
attrs:
  suffix:
  xsitype: proc:genProcData
  walltime: 24:00:00
  memory: 3850
  ppn: 4
  env: /share/apps/cmic/NiftyPipe/pCT_only/setup_pCT_only.sh
  type: scan
  scan_nb: scan1
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Functions used by dax_tools
"""

import csv
from datetime import datetime
import getpass
import glob
import imp
import json
import logging
import os
import shutil
import sys
from multiprocessing import Pool
import re

from . import bin
from . import XnatUtils
from . import utilities
from . import assessor_utils
from .dax_settings import (DAX_Settings, DAX_Netrc, DEFAULT_DATATYPE,
                           DEFAULT_FS_DATATYPE)
from .errors import (DaxUploadError, DaxError, DaxNetrcError, DaxSetupError)

from .task import (READY_TO_COMPLETE, COMPLETE, UPLOADING, JOB_FAILED,
                   JOB_PENDING, NEEDS_QA)
from .task import ClusterTask
from .XnatUtils import XnatUtilsError
from .version import VERSION as __version__
from .git_revision import git_revision as __git_revision__
from .suppdf import suppdf


# Global Variables
LOGGER = logging.getLogger('dax')

BASH_PROFILE_XNAT = """# XNAT host for dax:
{export_cmd}
"""

# Variables for upload
ERR_MSG = 'Error from XnatUtils when uploading: %s'
DAX_SETTINGS = DAX_Settings()
JOB_EXTENSION_FILE = DAX_SETTINGS.get_job_extension_file()
_COMPLETE_FLAG_FILE = 'READY_TO_COMPLETE.txt'
_READY_FLAG_FILE = 'READY_TO_UPLOAD.txt'
_FAILED_FLAG_FILE = 'JOB_FAILED.txt'
_EMAILED_FLAG_FILE = 'ALREADY_SEND_EMAIL.txt'
_OUTLOG = 'OUTLOG'
_TRASH = 'TRASH'
_PBS = 'PBS'
_FLAG_FILES = 'FlagFiles'
_UPLOAD_SKIP_LIST = [_OUTLOG, _TRASH, _PBS, _FLAG_FILES, 'TRIALSQ', 'SAVE', 'DISKQ']
SNAPSHOTS_ORIGINAL = 'snapshot_original.png'
SNAPSHOTS_PREVIEW = 'snapshot_preview.png'
DEFAULT_HEADER = ['host', 'username', 'password', 'projects']

# Cmd:
GS_CMD = """gs -q -o {original} -sDEVICE=pngalpha -dLastPage=1 {pdf_path}"""
CONVERT_CMD = """convert {original} -resize x200 {preview}"""
GS_CMD2 = """gs -q -o {original} -sDEVICE=pngalpha {pdf_path}"""

# WARNING content for emails
WARNING_START_CONTENT = """
The following assessors already exist and the Spider try to upload files on \
existing files :
"""
WARNING_END_CONTENT = """
You should:
    - remove the assessor if you want to upload the data
    - set the status of the assessor to "uploading"
    - remove the data from the upload folder if you do not want to upload.
"""
WARNING_SUBJECT = 'ERROR/WARNING: dax_upload'


def default_resdir():
    return os.path.join(
        '/scratch',
        os.environ.get('USER', 'USER'),
        'Spider_Upload_Dir')


def upload_tasks(logfile, debug, upload_settings=None,
                 host=None, username=None, password=None,
                 projects=None, suffix='', emailaddress=None,
                 uselocking=True, resdir=default_resdir(), num_threads=1):
    """
    Upload tasks from the queue folder.

    :param logfile: Full file of the file used to log to
    :param debug: Should debug mode be used
    :param upload_settings: settings file (csv, py, json) to define
                            xnat host/project relationship.
    :param host: XNAT host
    :param username: XNAT username
    :param password: XNAT password
    :param suffix: suffix for flagfile
    :param emailaddress: email address for warnings
    :param projects: Project(s) to upload

    """
    bin.set_logger(logfile, debug)

    FLAGFILE_TEMPLATE = os.path.join(resdir, _FLAG_FILES, 'Process_Upload_running')
    flagfile = "%s%s.txt" % (FLAGFILE_TEMPLATE, suffix)

    # Load the settings for upload
    upload_settings = load_upload_settings(upload_settings, host, username,
                                           password, projects)
    print_upload_settings(upload_settings, resdir)
    # create the flag file showing that the spider is running
    if uselocking and is_dax_upload_running(flagfile):
        return

    try:
        upload_results(upload_settings, emailaddress, resdir, num_threads)
    finally:
        if uselocking:
            # remove flagfile
            os.remove(flagfile)


def select_assessor(xnat, assessor_dict):
    """
    Select the assessor pyxnat Eobject from the assessor dictionary information

    :param xnat: pyxnat.interface object
    :param assessor_dict: assessor dictionary
    :return: assessor pyxnat Eobject
    """
    return xnat.select_assessor(
        assessor_dict['project_id'],
        assessor_dict['subject_label'],
        assessor_dict['session_label'],
        assessor_id=assessor_dict['label'])


def is_dax_upload_running(flagfile):
    """
    Check if dax_upload is not already running on the station

    :return: True if dax_upload already running, False otherwise.
    """
    if os.path.exists(flagfile):
        LOGGER.warn('Upload already running.')
        return True
    else:
        f_obj = open(flagfile, 'w')
        today = datetime.now()
        datestr = "Date: %s%s%s_%s:%s:%s" % (str(today.year),
                                             str(today.month),
                                             str(today.day),
                                             str(today.hour),
                                             str(today.minute),
                                             str(today.second))
        f_obj.write(datestr + '\n')
        f_obj.close()
        LOGGER.debug('Flagfile created: %s with date: %s'
                     % (flagfile, datestr))
        return False


def get_assessor_list(projects, resdir):
    """
    Get the list of assessors labels to upload to XNAT from the queue folder.

    :param projects: list of projects to upload to XNAT
    :return: list of assessor to upload from upload folder
    """
    assessor_label_list = list()

    LOGGER.debug(' - Get Processes names from the upload folder...')
    # check all files/folder in the directory
    dirs = list(filter(os.path.isdir,glob.glob(os.path.join(resdir, '*'))))
    dirs.sort(key=lambda x: os.path.getmtime(x))
    for assessor_label in dirs:
        assessor_label = os.path.basename(assessor_label)
        if assessor_label in _UPLOAD_SKIP_LIST:
            continue
        # If projects set, check that the project is in the list of projects
        # to upload to XNAT
        if projects and assessor_label.split('-x-')[0] not in projects:
            continue

        assessor_path = os.path.join(resdir, assessor_label)
        if not os.path.isdir(assessor_path):
            continue
        if os.path.exists(os.path.join(assessor_path, _EMAILED_FLAG_FILE)):
            LOGGER.debug('skipping, exists on XNAT:{}'.format(assessor_label))
            continue

        rflag = os.path.join(assessor_path, _READY_FLAG_FILE)
        fflag = os.path.join(assessor_path, _FAILED_FLAG_FILE)
        cflag = os.path.join(assessor_path, _COMPLETE_FLAG_FILE)

        if not (os.path.exists(rflag) or os.path.exists(fflag)):
            LOGGER.debug('skipping, still running:{}'.format(assessor_label))
            continue

        if not os.path.exists(cflag):
            LOGGER.debug('skipping, not completed:{}'.format(assessor_label))
            continue

        # Passed all checks, so add it to upload list
        assessor_label_list.append(assessor_label)

    return assessor_label_list


def get_pbs_list(projects, pbsdir):
    """
    Get the list of PBS file to upload to XNAT.

    :param projects: list of projects to upload to XNAT
    :return: list of pbs file from the PBS folder
    """
    pbs_list = list()

    LOGGER.debug(' - Get the PBS for the processes...')
    # check all files/folder in the directory
    for pbs_name in os.listdir(os.path.join(pbsdir)):
        # If projects set, check that the project is in the list of
        # projects to upload to XNAT
        if projects and pbs_name.split('-x-')[0] not in projects:
            continue

        pbs_file = os.path.join(pbsdir, pbs_name)
        if os.path.isfile(pbs_file):
            pbs_list.append(pbs_name)

    return pbs_list


def get_version_assessor(assessor_path):
    """
    Get the version of the assessor that we are uploading from text file

    :param assessor_path: path for the assessor
    :return: version of the assessor from the version.txt file
    """
    version = '1.0.0'
    if os.path.exists(os.path.join(assessor_path, 'version.txt')):
        f_obj = open(os.path.join(assessor_path, 'version.txt'), 'r')
        version = f_obj.read().strip()
        f_obj.close()
    return version


def get_dax_docker_version_assessor(assessor_path):
    """
    Get the dax_docker_version of assessor we are uploading from text file

    :param assessor_path: path for the assessor
    :return: version of the assessor from the version.txt file
    """
    dax_docker_version = ''
    fpath = os.path.join(assessor_path, 'dax_docker_version.txt')

    try:
        with open(fpath, 'r') as f_obj:
            dax_docker_version = f_obj.read().strip()

    except IOError as e:
        LOGGER.warn('failed to read dax_docker_version:' + str(e))

    return dax_docker_version


def generate_snapshots(assessor_path):
    """
    Generate Snapshots from the PDF if it exists.

    :param assessor_path: path for the assessor
    :return: None
    """
    snapshot_dir = os.path.join(assessor_path, 'SNAPSHOTS')
    snapshot_original = os.path.join(snapshot_dir, SNAPSHOTS_ORIGINAL)
    snapshot_preview = os.path.join(snapshot_dir, SNAPSHOTS_PREVIEW)
    try:
        pdf_path = glob.glob(assessor_path + '/PDF/*.pdf')[0]
    except Exception as err:
        LOGGER.debug('skipping generate_snapshots:{}:err={}'.format(assessor_path, err))
        return False

    if not os.path.exists(snapshot_original):
        LOGGER.debug('    +creating original of SNAPSHOTS')

        if not os.path.exists(snapshot_dir):
            os.mkdir(snapshot_dir)

        # Make the snapshots for the assessors with ghostscript
        cmd = GS_CMD.format(original=snapshot_original, pdf_path=pdf_path)
        LOGGER.debug(cmd)
        os.system(cmd)

    # Check for empty file
    if os.path.exists(snapshot_original) and os.stat(snapshot_original).st_size == 0:
        # Delete the empty file
        os.remove(snapshot_original)

        # Try the alternate ghostscript call
        cmd = GS_CMD2.format(original=snapshot_original, pdf_path=pdf_path)
        LOGGER.debug(cmd)
        os.system(cmd)

    # Create the preview snapshot from the original if Snapshots exist
    if os.path.exists(snapshot_original):
        # Make the snapshot_thumbnail
        LOGGER.debug('    +creating preview of SNAPSHOTS')
        cmd = CONVERT_CMD.format(
            original=snapshot_original, preview=snapshot_preview)
        LOGGER.debug(cmd)
        os.system(cmd)


def copy_outlog(assessor_dict, assessor_path, resdir):
    """
    Copy the oulog files to the assessor folder if we are uploading.

    :param assessor_dict: dictionary for the assessor
    :return: None
    """
    outlog_path = os.path.join(resdir, _OUTLOG,
                               assessor_dict['label'] + '.output')
    new_outlog_path = os.path.join(assessor_path, _OUTLOG,
                                   assessor_dict['label'] + '.output')
    if os.path.exists(outlog_path):
        os.makedirs(os.path.join(assessor_path, _OUTLOG))
        shutil.move(outlog_path, new_outlog_path)


def get_xsitype(assessor_dict):
    """
    Copy the outlog files to the assessor folder if we are uploading.

    :param assessor_dict: dictionary for the assessor
    :return: xsitype for the assessor_dict
    """
    proctype = assessor_dict['proctype']
    if proctype == 'FS':
        return DEFAULT_FS_DATATYPE
    else:
        return DEFAULT_DATATYPE


def is_complete(assessor_dict, assessor_path, procstatus):
    """
    Copy the oulog files to the assessor folder if we are uploading.

    :param assessor_dict: dictionary for the assessor
    :param procstatus: status to set for the assessor
    :return: True if the assessor is Complete, False otherwise
    """
    if procstatus == READY_TO_COMPLETE or procstatus == COMPLETE:
        eflag = os.path.join(assessor_path, _EMAILED_FLAG_FILE)
        open(eflag, 'w').close()
        LOGGER.warn('  -->Data already present on XNAT.\n')
        return True
    else:
        return False


def create_freesurfer_assessor(assessor_obj):
    """
    Create freesurfer specific assessor using the DEFAULT_FS_DATATYPE from dax

    :param assessor_obj: pyxnat assessor Eobject
    :return: None
    """
    # create the assessor and set the status
    assessor_obj.create(assessors=DEFAULT_FS_DATATYPE,
                        **{DEFAULT_FS_DATATYPE + '/fsversion': '0'})
    now = datetime.now()
    today = '%s-%s-%s-' % (str(now.year), str(now.month), str(now.day))

    assessor_obj.attrs.mset(
        {DEFAULT_FS_DATATYPE + '/validation/status': JOB_PENDING,
         DEFAULT_FS_DATATYPE + '/date': today})


def create_default_assessor(assessor_obj, proctype):
    """
    Create default assessor using the DEFAULT_DATATYPE from dax

    :param assessor_obj: pyxnat assessor Eobject
    :param proctype: proctype for the assessor
    :return: None
    """
    # Create the assessor and set attributes
    now = datetime.now()
    today = '%s-%s-%s-' % (str(now.year), str(now.month), str(now.day))

    assessor_obj.create(assessors=DEFAULT_DATATYPE)
    # Call mset to only make a single HTTP request
    assessor_obj.attrs.mset(
        {DEFAULT_DATATYPE + '/validation/status': JOB_PENDING,
         DEFAULT_DATATYPE + '/proctype': proctype,
         DEFAULT_DATATYPE + '/date': today})


def should_upload_assessor(assessor_obj,
                           assessor_dict, assessor_path, version):
    """
    Check if the assessor is ready to be uploaded to XNAT

    :param assessor_obj: pyxnat assessor Eobject
    :param assessor_dict: assessor dictionary
    :param xsitype: xsitype for the assessor (fsData or proc:GenProcData, ...)
    :param version: version for the assessor
    :return: True if the assessor should be upload, False otherwise
    """
    if not assessor_obj.exists():
        return False
        # if xsitype == DEFAULT_FS_DATATYPE:
        #     create_freesurfer_assessor(assessor_obj)
        # else:
        #     create_default_assessor(assessor_obj, assessor_dict['proctype'])
    else:
        xsitype = assessor_obj.datatype()
        # Check if not already complete assessor
        procstatus = assessor_obj.attrs.get(xsitype + '/procstatus')
        if is_complete(assessor_dict, assessor_path, procstatus):
            return False
    # set the status to UPLOADING
    assessor_obj.attrs.mset({xsitype + '/procstatus': UPLOADING,
                             xsitype + '/procversion': version})
    return True


def upload_assessor(xnat, assessor_dict, assessor_path, resdir):
    """
    Upload results to an assessor

    :param xnat: pyxnat.Interface object
    :param assessor_dict: assessor dictionary
    :return: None
    """
    # get spiderpath from version.txt file:
    diskq_dir = os.path.join(resdir, 'DISKQ')
    version = get_version_assessor(assessor_path)
    dax_docker_version = get_dax_docker_version_assessor(assessor_path)
    session_obj = xnat.select_session(
        assessor_dict['project_id'],
        assessor_dict['subject_label'],
        assessor_dict['session_label'])

    if not session_obj.exists():
        LOGGER.error('Cannot upload assessor, session does not exist.')
        return True

    # Select assessor
    assessor_dict = assessor_utils.parse_full_assessor_name(
            os.path.basename(assessor_path))
    assessor_obj = session_obj.assessor(assessor_dict['label'])
    if should_upload_assessor(assessor_obj,
                              assessor_dict,
                              assessor_path,
                              version):
        xsitype = assessor_obj.datatype()
        # Before Upload
        LOGGER.debug('suppdf')
        suppdf(assessor_path, assessor_obj)
        LOGGER.debug('generate_snapshots')
        generate_snapshots(assessor_path)
        LOGGER.debug('copy_outlog')
        copy_outlog(assessor_dict, assessor_path, resdir)

        # Upload the XML if FreeSurfer
        if xsitype == DEFAULT_FS_DATATYPE:
            xmlpath = os.path.join(assessor_path, 'XML')
            if os.path.exists(xmlpath):
                LOGGER.debug('    +setting XML for FreeSurfer')
                xml_files_list = os.listdir(xmlpath)
                if len(xml_files_list) != 1:
                    fpath = assessor_path
                    msg = 'cannot upload FreeSurfer assessor, \
unable to find XML file: %s'
                    LOGGER.error(msg % (fpath))
                    return
                xml_path = os.path.join(assessor_path, 'XML',
                                        xml_files_list[0])
                assessor_obj.create(xml=xml_path, allowDataDeletion=False)

        # Upload
        # for each folder=resource in the assessor directory
        for resource in os.listdir(assessor_path):
            resource_path = os.path.join(assessor_path, resource)
            # Need to be in a folder to create the resource :
            if os.path.isdir(resource_path):
                LOGGER.debug('    +uploading %s' % (resource))
                try:
                    upload_resource(assessor_obj, resource, resource_path)
                except Exception as e:
                    try:
                        LOGGER.warn('failed to upload, trying again')
                        upload_resource(assessor_obj, resource, resource_path)
                    except Exception as e:
                        import traceback
                        LOGGER.error(traceback.format_exc())
                        _msg = 'failed to upload, skipping assessor:{}:{}'.format(
                            resource_path, str(e))
                        LOGGER.error(_msg)
                        return

        # after Upload
        if is_diskq_assessor(os.path.basename(assessor_path), resdir):
            # was this run using the DISKQ option
            # Read attributes
            ctask = ClusterTask(assessor_dict['label'], resdir, diskq_dir)

            # Set on XNAT
            assessor_obj.attrs.mset({
                xsitype + '/procstatus': ctask.get_status(),
                xsitype + '/validation/status': NEEDS_QA,
                xsitype + '/jobid': ctask.get_jobid(),
                xsitype + '/jobnode': ctask.get_jobnode(),
                xsitype + '/memused': ctask.get_memused(),
                xsitype + '/walltimeused': ctask.get_walltime(),
                xsitype + '/jobstartdate': ctask.get_jobstartdate(),
                xsitype + '/dax_version': __version__,
                xsitype + '/dax_version_hash': __git_revision__,
                xsitype + '/dax_docker_version': dax_docker_version
            })

            # Delete the task from diskq
            ctask.delete()
        elif os.path.exists(os.path.join(assessor_path, _READY_FLAG_FILE)):
            assessor_obj.attrs.set(xsitype + '/procstatus', READY_TO_COMPLETE)
        else:
            assessor_obj.attrs.set(xsitype + '/procstatus', JOB_FAILED)

        # Remove the folder
        shutil.rmtree(assessor_path)

        return True

    return False


def is_diskq_assessor(assr_label, resdir):
    # Does a batch file exist for this assessor?
    afile = os.path.join(resdir, 'DISKQ', 'BATCH', assr_label + JOB_EXTENSION_FILE)
    return os.path.exists(afile)


def upload_resource(assessor_obj, resource, resource_path):
    """
    Upload a resource folder to an assessor

    :param assessor_obj: pyxnat assessor Eobject
    :param resource: resource to upload
    :param resource_path: resource path on the station
    :return: None
    """
    if resource == 'SNAPSHOTS':
        upload_snapshots(assessor_obj, resource_path)
    else:
        rfiles_list = os.listdir(resource_path)
        if not rfiles_list:
            LOGGER.warn('No files in {}'.format(resource_path))
        elif len(rfiles_list) > 1 or os.path.isdir(os.path.join(resource_path, rfiles_list[0])):
            try:
                XnatUtils.upload_folder_to_obj(
                    resource_path, assessor_obj.out_resource(resource),
                    resource, removeall=True)
            except XnatUtilsError as err:
                print((ERR_MSG % err))
        else:
            # One file, just upload it
            fpath = os.path.join(resource_path, rfiles_list[0])
            try:
                XnatUtils.upload_file_to_obj(
                    fpath, assessor_obj.out_resource(resource), removeall=True)
            except XnatUtilsError as err:
                print((ERR_MSG % err))


def upload_snapshots(assessor_obj, resource_path):
    """
    Upload snapshots to an assessor

    :param assessor_obj: pyxnat assessor Eobject
    :param resource_path: resource path on the station
    :return: None
    """
    # Remove the previous Snapshots:
    if assessor_obj.out_resource('SNAPSHOTS').exists:
        assessor_obj.out_resource('SNAPSHOTS').delete()
    original = os.path.join(resource_path, SNAPSHOTS_ORIGINAL)
    thumbnail = os.path.join(resource_path, SNAPSHOTS_PREVIEW)
    status = None
    try:
        status = XnatUtils.upload_assessor_snapshots(
            assessor_obj, original, thumbnail)
    except XnatUtilsError as err:
        print((ERR_MSG % err))

    if status:
        os.remove(original)
        os.remove(thumbnail)
    else:
        LOGGER.warn('No snapshots original or preview were uploaded')

    # Upload the rest of the files in snapshots
    if len(os.listdir(resource_path)) > 0:
        try:
            XnatUtils.upload_folder_to_obj(
                resource_path, assessor_obj.out_resource('SNAPSHOTS'),
                'SNAPSHOTS')
        except XnatUtilsError as err:
            print((ERR_MSG % err))


def upload_assessors(xnat, projects, resdir, num_threads=1):
    """
    Upload all assessors to XNAT

    :param xnat: pyxnat.Interface object
    :param projects: list of projects to upload to XNAT
    :return: None
    """
    # Get the assessor label from the directory :
    assessors_list = get_assessor_list(projects, resdir)
    number_of_processes = len(assessors_list)
    warnings = list()

    LOGGER.info(('Starting upload pool:{} threads'.format(str(num_threads))))
    sys.stdout.flush()

    pool = Pool(processes=num_threads)
    for index, assessor_label in enumerate(assessors_list):
        LOGGER.info(index)
        sys.stdout.flush()

        pool.apply_async(
            upload_thread,
            [xnat, index, assessor_label, number_of_processes, resdir])

    LOGGER.info('waiting for upload pool to finish...')
    sys.stdout.flush()

    pool.close()
    pool.join()

    LOGGER.info('upload pool finished')
    sys.stdout.flush()

    return warnings


def upload_resource_subjgenproc(sassr, resource, dirpath):
    # TODO: merge this with the other upload_resource

    if resource == 'SNAPSHOTS':
        upload_snapshots_subjgenproc(sassr, dirpath)
    else:
        # Get list of files in the dir
        rfiles_list = os.listdir(dirpath)

        # Handle none, multiple or one file found
        if not rfiles_list:
            LOGGER.warn('No files in {}'.format(dirpath))
        elif len(rfiles_list) > 1 or os.path.isdir(rfiles_list[0]):
            # Upload the whole dir
            XnatUtils.upload_folder_to_obj(
                dirpath,
                sassr.resource(resource),
                resource,
                removeall=True)
        else:
            # One file, upload it
            fpath = os.path.join(dirpath, rfiles_list[0])
            XnatUtils.upload_file_to_obj(
                fpath,
                sassr.resource(resource),
                removeall=True)


def upload_snapshots_subjgenproc(sassr, dirpath):
    # TODO: merge this with the other upload_snapshots
    """
    Upload snapshots to an assessor
    :param assessor_obj: pyxnat assessor Eobject
    :param dirpath: local path
    :return: None
    """
    orig = os.path.join(dirpath, SNAPSHOTS_ORIGINAL)
    thumb = os.path.join(dirpath, SNAPSHOTS_PREVIEW)

    if not os.path.isfile(orig):
        LOGGER.warn('original snapshot not found:{}'.format(orig))
    else:
        # Upload the full size image
        sassr.resource('SNAPSHOTS').file(os.path.basename(orig)).put(
            orig,
            orig.split('.')[1].upper(),
            'ORIGINAL',
            overwrite=True,
            params={"event_reason": "DAX upload"})

    if not os.path.isfile(thumb):
        LOGGER.warn('thumbnail snapshot not found:{}'.format(thumb))
    else:
        # Upload the thumbnail image
        sassr.resource('SNAPSHOTS').file(os.path.basename(thumb)).put(
            thumb,
            thumb.split('.')[1].upper(),
            'THUMBNAIL',
            overwrite=True,
            params={"event_reason": "DAX upload"})


def upload_assessor_subjgenproc(xnat, dirpath):
    resdir = os.path.dirname(dirpath)
    diskq_dir = os.path.join(resdir, 'DISKQ')
    assr = os.path.basename(dirpath)
    [proj, subj, proctype, guid] = assr.split('-x-')

    # Get the assessor object on XNAT
    LOGGER.debug('connecting to assessor on xnat')

    sassr = xnat.select('/projects/{}/subjects/{}/experiments/{}'.format(
        proj, subj, assr))

    if not sassr.exists():
        # LOGGER.info('creating subject asssessor'.format(assr))
        # sassr.create(experiments='proc:subjgenprocdata')
        # We will assume the assessor has already been made elsewhere
        # with the inputs field complete as well
        # as proctype, procversion, and date
        LOGGER.info('assessor does not exist, refusing to create')
        return

    # Double-check the status
    procstatus = sassr.attrs.get(sassr.datatype() + '/procstatus')
    if procstatus in [READY_TO_COMPLETE, COMPLETE]:
        # LOGGER.info('creating subject asssessor'.format(assr))
        # sassr.create(experiments='proc:subjgenprocdata')
        # We will assume the assessor has already been made elsewhere
        # with the inputs field complete as well
        # as proctype, procversion, and date
        LOGGER.info('assessor complete, refusing to overwite')
        return

    LOGGER.info('uploading:{}'.format(dirpath))

    # Before Upload
    LOGGER.debug('suppdf')
    suppdf(dirpath, sassr)
    LOGGER.debug('generate_snapshots')
    generate_snapshots(dirpath)
    # note, we don't need to copy outlog, we assume it's been done during
    # update()

    # Upload each resource
    for resource in os.listdir(dirpath):
        resource_path = os.path.join(dirpath, resource)
        if not os.path.isdir(resource_path):
            LOGGER.debug('skipping, not a directory:{}'.format(resource_path))
            continue

        LOGGER.debug('+uploading:{}'.format(resource))
        try:
            upload_resource_subjgenproc(sassr, resource, resource_path)
        except Exception as e:
            _msg = 'upload failed, skipping assessor:{}:{}'.format(
                resource_path, str(e))
            LOGGER.error(_msg)

    # after Upload
    ctask = ClusterTask(assr, resdir, diskq_dir)

    # Set on XNAT
    _docker_version = get_dax_docker_version_assessor(dirpath) or 'null'
    _status = ctask.get_status() or 'null'
    _jobid = ctask.get_jobid() or 'null'
    _jobnode = ctask.get_jobnode() or 'null'
    _memused = ctask.get_memused() or 'null'
    _walltime = ctask.get_walltime() or 'null'
    _jobstartdate = ctask.get_jobstartdate() or 'null'
    LOGGER.debug('setting attributes on xnat')
    sassr.attrs.mset({
        'proc:subjgenprocdata/procstatus': _status,
        'proc:subjgenprocdata/validation/status': NEEDS_QA,
        'proc:subjgenprocdata/jobid': _jobid,
        'proc:subjgenprocdata/jobnode': _jobnode,
        'proc:subjgenprocdata/memused': _memused,
        'proc:subjgenprocdata/walltimeused': _walltime,
        'proc:subjgenprocdata/dax_docker_version': _docker_version,
        'proc:subjgenprocdata/jobstartdate': _jobstartdate,
        'proc:subjgenprocdata/dax_version': __version__,
        'proc:subjgenprocdata/dax_version_hash': __git_revision__,
    })

    # Delete the task from diskq
    ctask.delete()

    # Remove the folder
    shutil.rmtree(dirpath)


def upload_thread(xnat, index, assessor_label, number_of_processes, resdir):
    assessor_path = os.path.join(resdir, assessor_label)
    msg = "    *Process: %s/%s -- label: %s / time: %s"
    LOGGER.info(msg % (str(index + 1),str(number_of_processes), assessor_label, str(datetime.now())))

    if assessor_utils.is_sgp_assessor(assessor_label):
        uploaded = upload_assessor_subjgenproc(
            xnat, assessor_path)
    else:
        assessor_dict = assessor_utils.parse_full_assessor_name(assessor_label)
        if assessor_dict:
            uploaded = upload_assessor(
                xnat, assessor_dict, assessor_path, resdir)
            if not uploaded:
                mess = """    - Assessor label : {label}\n"""
                LOGGER.warn(mess.format(label=assessor_dict['label']))
        else:
            LOGGER.warn('     --> wrong label')

    # Disconnecting here because each thread actually acquires
    # a unique JSESSIONOID, not sure why yet or if this is what we want
    # but have confirmed this is what is happening - bdb 2020-01-28
    xnat.disconnect()


def upload_pbs(xnat, projects, resdir):
    """
    Upload all pbs files to XNAT

    :param xnat: pyxnat.Interface object
    :param projects: list of projects to upload to XNAT
    :return: None
    """
    pbsdir = os.path.join(resdir, _PBS)
    pbs_list = get_pbs_list(projects, pbsdir)
    number_pbs = len(pbs_list)
    for index, pbsfile in enumerate(pbs_list):
        pbs_fpath = os.path.join(pbsdir, pbsfile)
        mess = """   *Uploading PBS {index}/{max} -- File name: {file}"""
        LOGGER.info(mess.format(index=str(index + 1),
                                max=str(number_pbs),
                                file=pbsfile))
        assessor_label = os.path.splitext(pbsfile)[0]
        assessor_dict = assessor_utils.parse_full_assessor_name(assessor_label)
        if not assessor_dict:
            LOGGER.warn('wrong assessor label for %s' % (pbsfile))
            os.rename(pbs_fpath, os.path.join(resdir, _TRASH, pbsfile))
        else:
            assessor_obj = select_assessor(xnat, assessor_dict)
            if not assessor_obj.exists():
                LOGGER.warn('assessor does not exist for %s' % (pbsfile))
                new_location = os.path.join(resdir, _TRASH, pbsfile)
                os.rename(pbs_fpath, new_location)
            else:
                resource_obj = assessor_obj.out_resource(_PBS)
                if resource_obj.exists():
                    label = assessor_dict['label']
                    msg = 'the PBS resource already exists for the assessor %s'
                    LOGGER.warn(msg % (label))
                    adir = os.path.join(resdir, assessor_dict['label'])
                    if os.path.isdir(adir):
                        msg = 'Copying the pbs file in the assessor folder...'
                        LOGGER.warn(msg)
                        pbs_folder = os.path.join(adir, _PBS)
                        if not os.path.exists(pbs_folder):
                            os.mkdir(pbs_folder)
                        os.rename(pbs_fpath, os.path.join(pbs_folder, pbsfile))
                    else:
                        LOGGER.warn('Copying the pbs file in the TRASH ...')
                        trash = os.path.join(resdir, _TRASH, pbsfile)
                        os.rename(pbs_fpath, trash)
                else:
                    # upload the file
                    try:
                        status = XnatUtils.upload_file_to_obj(pbs_fpath,
                                                              resource_obj)
                    except XnatUtilsError as err:
                        LOGGER.error((ERR_MSG % err))
                    if status:
                        os.remove(pbs_fpath)


def upload_outlog(xnat, projects, resdir):
    """
    Upload all outlog files to XNAT

    :param xnat: pyxnat.Interface object
    :param projects: list of projects to upload to XNAT
    :return: None
    """
    outlogs_list = os.listdir(os.path.join(resdir, _OUTLOG))
    if projects:
        outlogs_list = [logfile for logfile in outlogs_list
                        if logfile.split('-x-')[0] in projects]

    number_outlog = len(outlogs_list)
    for index, outlogfile in enumerate(outlogs_list):
        outlog_fpath = os.path.join(resdir, _OUTLOG, outlogfile)
        mess = """   *Checking OUTLOG {index}/{max} -- File name: {file}"""
        LOGGER.info(mess.format(index=str(index + 1),
                                max=str(number_outlog),
                                file=outlogfile))
        assessor_label = os.path.splitext(outlogfile)[0]
        assessor_dict = assessor_utils.parse_full_assessor_name(assessor_label)
        if not assessor_dict:
            LOGGER.warn('     wrong outlog file. You should remove it')
        else:
            assessor_obj = select_assessor(xnat, assessor_dict)
            if not assessor_obj.exists():
                msg = '     no assessor on XNAT -- moving file to trash.'
                LOGGER.warn(msg)
                new_location = os.path.join(resdir, _TRASH, outlogfile)
                os.rename(outlog_fpath, new_location)
            else:
                if assessor_obj.attrs.get(
                        assessor_obj.datatype() + '/procstatus') == JOB_FAILED:
                    resource_obj = assessor_obj.out_resource(_OUTLOG)
                    if resource_obj.exists():
                        pass
                    else:
                        LOGGER.info('     uploading file.')
                        try:
                            status = XnatUtils.upload_file_to_obj(outlog_fpath,
                                                                  resource_obj)
                        except XnatUtilsError as err:
                            print((ERR_MSG % err))
                        if status:
                            os.remove(outlog_fpath)


def upload_results(upload_settings, emailaddress, resdir, num_threads=1):
    """
    Main function to upload the results / PBS / OUTLOG of assessors
     from the queue folder

    :param upload_settings: dictionary defining the upload information
    :return: None
    """
    # TODO: move resdir into upload_sttings
    if len(os.listdir(resdir)) == 0:
        LOGGER.warn('No data need to be uploaded.\n')
        sys.exit()

    warnings = list()

    for upload_dict in upload_settings:
        try:
            with XnatUtils.get_interface(host=upload_dict['host'],
                                         user=upload_dict['username'],
                                         pwd=upload_dict['password']) as intf:
                LOGGER.info('=' * 50)
                proj_str = (upload_dict['projects'] if upload_dict['projects']
                            else 'all')
                LOGGER.info('Connecting to XNAT <%s>, upload for projects:%s' %
                            (upload_dict['host'], proj_str))
                if not XnatUtils.has_dax_datatypes(intf):
                    msg = 'dax datatypes are not installed on xnat <%s>.'
                    raise DaxUploadError(msg % (upload_dict['host']))

                # 1) Upload the assessor data
                # For each assessor label that need to be upload :
                LOGGER.info('Uploading results for assessors')

                warnings.extend(
                    upload_assessors(intf, upload_dict['projects'], resdir,
                                     num_threads))

                # 2) Upload the PBS files
                # For each file, upload it to the PBS resource
                LOGGER.info('Uploading PBS files ...')
                upload_pbs(intf, upload_dict['projects'], resdir)

                # 3) Upload the OUTLOG files not uploaded with processes
                LOGGER.info('Checking OUTLOG files for JOB_FAILED jobs ...')
                upload_outlog(intf, upload_dict['projects'], resdir)
        except DaxNetrcError as e:
            msg = e.msg
            LOGGER.error(e.msg)


def load_upload_settings(f_settings, host, username, password, projects):
    """
    Function to parse arguments base on argparse

    :param f_settings: file to define the settings for uploading
    :param host: XNAT host
    :param username: XNAT username
    :param password: XNAT password
    :param projects: XNAT list of projects

    :return: list of dictionaries info_dict
       info_dict for the host [key:value]:
        host : string for XNAT host
        username : string for XNAT username
        password : string for XNAT password
          (can be the environment variable containing the value)
        projects : list of projects to upload for the host
    """
    host_projs = list()
    # If settings file given, load it and use it:
    if f_settings is not None:
        up_file = os.path.abspath(f_settings)
        if not os.path.isfile(up_file):
            raise DaxError('No upload settings file found: %s' % up_file)
        if f_settings.endswith('.json'):
            with open(up_file) as data_file:
                host_projs = json.load(data_file)
        elif f_settings.endswith('.py'):
            settings = imp.load_source('settings', up_file)
            host_projs = settings.host_projects
        elif f_settings.endswith('.csv'):
            with open(up_file, 'rb') as csvfileread:
                csvreader = csv.reader(csvfileread, delimiter=',')
                for index, row in (csvreader):
                    if len(row) < 4:
                        raise DaxError("error: could not read the csv row. \
Missing args. 4 needed, %s found at line %s." % (str(len(row)), str(index)))
                    else:
                        if row != DEFAULT_HEADER:
                            host_projs.append(dict(list(zip(DEFAULT_HEADER,
                                                        row[:4]))))
        elif f_settings.endswith('.yaml'):
            doc = utilities.read_yaml(f_settings)
            host_projs = doc.get('settings')
        else:
            raise DaxError("error: doesn't recognize the file format for the \
settings file. Please use either JSON/PYTHON/CSV format.")
    else:  # if not file, use the environment variables and options
        _host = os.environ['XNAT_HOST']
        if host:
            _host = host
        if projects:
            projects = projects.split(',')
        else:
            projects = []
        if username:
            username = username
            if not password:
                MSG = "Please provide the password for user <%s> on xnat(%s):"
                password = getpass.getpass(prompt=MSG % (username, _host))
                if not password:
                    raise DaxError('empty password entered')
        else:
            netrc_obj = DAX_Netrc()
            username, password = netrc_obj.get_login(_host)
        host_projs.append(dict(list(zip(DEFAULT_HEADER, [_host, username,
                                                         password,
                                                         projects]))))
    return host_projs


def print_upload_settings(upload_settings, resdir):
    """
    Display Host/Username/Projects that will be used to upload data from
     the queue.

    :return: None
    """
    LOGGER.info('Upload Settings selected by user:')
    for info in upload_settings:
        proj_str = ','.join(info['projects']) if info['projects'] else 'all'
        user_str = info['username'] if info['username'] else ''
        msg = 'XNAT Host: %s -- Xnat Username: %s -- projects: %s'
        LOGGER.info(msg % (info['host'], user_str, proj_str))
    LOGGER.info('Upload Directory: %s ' % (resdir))


def setup_dax_package():
    """ Setup dax package """
    print('########## DAX_SETUP ##########')

    # Set xnat credentials if needed
    set_xnat_netrc()

    print('########## END ##########')


def test_connection_xnat(host, user, pwd):
    """
    Method to check connection to XNAT using host, user, pwd.

    :param host: Host for XNAT
    :param user: User for XNAT
    :param pwd: Password for XNAT
    :return: True if succeeded, False otherwise.
    """
    from pyxnat.core.errors import DatabaseError
    from pyxnat import Interface
    try:
        xnat = Interface(host, user, pwd)
        # try deleting SESSION connection
        xnat._exec('/data/JSESSION', method='DELETE')
        print(' --> Good login.')
        return True
    except DatabaseError:
        print(' --> error: Wrong login.')
        return False


def set_xnat_netrc():
    """Ask User for xnat credentials and store it in xnatnetrc file.

    :return: None
    """
    netrc_obj = DAX_Netrc()
    if netrc_obj.is_empty():
        LOGGER.warn('Setting XNAT login, netrc is empty.')
        connection = False
        while not connection:
            host = input("Please enter your XNAT host: ")
            user = input("Please enter your XNAT username: ")
            pwd = getpass.getpass(prompt='Please enter your XNAT password: ')
            connection = test_connection_xnat(host, user, pwd)

        # Add host to your netrc
        netrc_obj.add_host(host, user, pwd)

        # add XNAT_HOST to your profile file:
        init_profile(host)
    else:
        print('dax setup is already complete')


def init_profile(host):
    """Function to init your profile file to call xnat_profile.

    :param host: Host of XNAT to add to your profile
    :return: None
    """
    # link the file in the bashrc or profile
    profile = os.path.join(os.path.expanduser('~'), '.bash_profile')

    if os.path.exists(os.path.join(os.path.expanduser('~'), '.bash_profile')):
        profile = os.path.join(os.path.expanduser('~'), '.bash_profile')
    elif os.path.exists(os.path.join(os.path.expanduser('~'), '.bashrc')):
        profile = os.path.join(os.path.expanduser('~'), '.bashrc')
    elif os.path.exists(os.path.join(os.path.expanduser('~'), '.profile')):
        profile = os.path.join(os.path.expanduser('~'), '.profile')
    else:
        raise DaxSetupError("could not find your profile file.")

    # Add the line to the profile
    if 'XNAT_HOST' not in open(profile).read():
        print('Adding XNAT_HOST to your profile:{}'.format(profile))
        line_to_add = 'export XNAT_HOST=%s' % host
        with open(profile, "a") as f_profile:
            f_profile.write(BASH_PROFILE_XNAT.format(export_cmd=line_to_add))
    else:
        print('XNAT_HOST already in profile:{}'.format(profile))



'''
BIDS to XNAT

Extract the information from BIDS for Xnatupload

@author: Praitayini Kanakaraj, Electrical Engineering, Vanderbilt University

'''
import os
import sys
import json
import glob

# check if valid bids
# extract and map to dict.


def split_name_with_nii(filename):
    """
    Returns the clean basename and extension of a file.
    Means that this correctly manages the ".nii.gz" extensions.

    Parameters
    ----------
    filename: str
        The filename to clean

    Returns
    -------
        base, ext : tuple(str, str)
        Clean basename and the full extension
    """
    base, ext = os.path.splitext(filename)

    if ext == ".gz":
        # Test if we have a .nii additional extension
        temp_base, add_ext = os.path.splitext(base)

        if add_ext == ".nii":
            ext = add_ext + ext
            base = temp_base

    return base, ext


def transform_to_xnat(bids_dir, project):
    """
    Method to transfrom bids to xnat

    :param bids_dir: BIDS Directory
    :param project: XNAT project ID
    :return: uplaod_scan -> the dict with info of scans to upload
    """
    # Check bids dir path exists
    if not os.path.exists(bids_dir):
        print('ERROR: %s path does not exists' % (bids_dir))
        exit()

    # Extract the values from the bids data
    bids_dict = {}
    upload_scan = []
    unq_scan_id = 1
    pre_dir = None
    xnat_dataset = dataset_source_xnat(bids_dir)
    diffusion_dict = {}
    for root, dirs, files in os.walk(bids_dir):

        # increment scan id if the previous dir is different
        cur_dir = root.rsplit('/', 1)[0]
        if pre_dir is not None and cur_dir != pre_dir:
            unq_scan_id = 1
        pre_dir = cur_dir

        # Sorting the files helps with DWI files, files will be
        # grouped together in this order: bval/bvec/nii.gz
        for i in sorted(files):
            if i.endswith('nii.gz') or i.endswith('.bvec') \
                    or i.endswith('.bval'):
                bids_filename_contents = i.split('_')

                if xnat_dataset:
                    # get info from json file for data dervied from XNAT
                    bids_filename = i.split('.')[0]
                    json_file = bids_filename + '.json'
                    with open(os.path.join(root, json_file), 'r') as f:
                        json_contents = json.load(f)

                    # subj from json
                    subject = json_contents['XNATProvenance'].split('/')[8]
                    bids_dict['subject_label'] = subject

                    # sess from json
                    session = json_contents['XNATProvenance'].split('/')[10]
                    bids_dict['session_label'] = session

                    # series des (on xnat/bidsmap) from bids
                    bids_dict['series_description'] = json_contents['SeriesDescription']

                    # label <project>-x-<subject>-x-<session>-x-<ID>
                    scan_id = json_contents['XNATProvenance'].split('/')[12]
                    bids_dict['label'] = '-'.join((project,
                                                  subject, session, scan_id))

                    # type quality from json
                    bids_dict['ID'] = scan_id
                    bids_dict['type'] = json_contents['ScanType']
                    bids_dict['quality'] = json_contents['ScanQuality']

                    # resource and resource path
                    bids_filepath = os.path.join(root, i)

                    if bids_filepath.endswith('nii.gz'):
                        bids_dict['resource'] = {'NIFTI': [bids_filepath]}
                    if bids_filepath.endswith('bvec'):
                        bids_dict['resource'] = {'BVEC': [bids_filepath]}
                    if bids_filepath.endswith('bval'):
                        bids_dict['resource'] = {'BVAL': [bids_filepath]}

                else:
                    # get data from filename for public bids dataset
                    # sub, sess from bids
                    subject = [(i.split('-')[1])
                               for i in bids_filename_contents if i.startswith('sub')][0]
                    bids_dict['subject_label'] = subject
                    try:
                        session = [(i.split('-')[1])
                                   for i in bids_filename_contents if i.startswith('ses')][0]
                    except IndexError:
                        session = subject
                    # xnatupload needs unique session id
                    bids_dict['session_label'] = subject + '-' + session

                    basename, ext = split_name_with_nii(i)
                    if ext == '.bval' and basename not in diffusion_dict:
                        diffusion_dict[basename] = unq_scan_id
                    elif basename in diffusion_dict:
                        unq_scan_id = diffusion_dict[basename]
                    # id increment unique value
                    bids_dict['ID'] = "{0:0=2d}".format(unq_scan_id)

                    # label from bids datatype
                    datatype = root.split('/')[-1]
                    scan_id = datatype
                    bids_dict['label'] = '-'.join((project,
                                                  subject, session, scan_id))

                    # series_des type from last key in bids + run + acq
                    try:
                        run_number = [(i.split('-')[1])
                                      for i in bids_filename_contents if i.startswith('run')][0]
                    except IndexError:
                        run_number = ''

                    bids_dict['series_description'] = i.split(
                        '.')[0].split('_')[-1] + run_number
                    bids_dict['type'] = i.split(
                        '.')[0].split('_')[-1] + run_number

                    bids_dict['quality'] = 'questionable'

                    bids_filepath = os.path.join(root, i)
                    if bids_filepath.endswith('nii.gz'):
                        bids_dict['resource'] = {'NIFTI': [bids_filepath]}
                    if bids_filepath.endswith('bvec'):
                        bids_dict['resource'] = {'BVEC': [bids_filepath]}
                    if bids_filepath.endswith('bval'):
                        bids_dict['resource'] = {'BVAL': [bids_filepath]}
                    unq_scan_id += 1

                # other keys
                bids_dict['object_type'] = 'scan'
                bids_dict['project_id'] = project
                # check the 4 datatypes for MR
                if bids_filepath.split('/')[-2] == 'anat' or 'func' or 'dwi' \
                        or 'fmap' or 'perf':
                    bids_dict['session_type'] = 'MR'
                else:
                    sys.exit()

                upload_scan.append(bids_dict.copy())

    return upload_scan


def dataset_source_xnat(bids_dir):
    """
    Method to check if the data was downloaded from xnat

    :param bids_dir: BIDS Directory
    :return: True or False
    """
    dataset_description_file = glob.glob(
        bids_dir + "/**/dataset_description.json", recursive=True)
    if not os.path.exists(dataset_description_file[0]):
        return False
    else:
        with open(dataset_description_file[0], 'r') as f:
            json_contents = json.load(f)
            if 'DatasetDOI' not in json_contents:
                return False
            elif not json_contents['DatasetDOI'].endswith('xnat'):
                return False
    return True
    


""" Module classes for Scan and Sessions """

from datetime import datetime
import logging
import os
import shutil

from . import XnatUtils
from .dax_settings import DAX_Settings
from . import utilities


__copyright__ = 'Copyright 2013 Vanderbilt University. All Rights Reserved'
__all__ = ['Module', 'ScanModule', 'SessionModule']
DAX_SETTINGS = DAX_Settings()
# Logger for logs
LOGGER = logging.getLogger('dax')


class Module(object):
    """ Object Module to create a module for DAX
        Module runs directly during a build on a session or scan
        to generate inputs data for scans/sessions
    """
    def __init__(self, mod_name, directory, email, text_report, smtp_host=None):
        """
        Entry point of the Base Module Class.

        :param mod_name: Name of the module
        :param directory: Temp directory to store data
        :param email: email address to send report
        :param text_report: string to write for report email
        :return: None
        """
        self.mod_name = mod_name
        self.directory = directory
        self.email = XnatUtils.get_input_list(input_val=email,
                                              default_val=None)
        self.text_report = text_report
        self.send_an_email = 0
        self.smtp_host = smtp_host

    def needs_run(self):
        """
        Check if the module needs to run
        Implemented in derived classes.

        :return: True if it does, False otherwise
        """
        raise NotImplementedError()

    def prerun(self):
        """
        Method to run before looping on a project
        Implemented in derived classes.

        :return: None
        """
        raise NotImplementedError()

    def afterrun(self):
        """
        Method to run after looping on a project
        Implemented in derived classes.

        :return: None
        """
        raise NotImplementedError()

    def report(self, string):
        """
        Add report to an email and send it at the end of the module

        :param string: string to add to the email
        :return: None
        """
        self.text_report += """  -{content}\n""".format(content=string)
        self.send_an_email = 1

    def get_report(self):
        """
        Get the report text

        :return: text_report variable
        """
        return self.text_report

    def make_dir(self, suffix=''):
        """
        Create the tmp directory for the modules

        :param suffix: suffix to add to the directory name
        :return: None
        """
        # add the suffix if one to the directory:
        if suffix:
            if suffix not in self.directory:
                self.directory = '%s_%s' % (self.directory.rstrip('/'), suffix)

        # Check if the directory exists
        if not os.path.exists(self.directory):
            os.mkdir(self.directory)
        else:
            if suffix:
                self.clean_directory()
            else:
                today = datetime.now()
                dir_format = "%s_tmp_%s_%s_%s_%s_%s_%s"""
                fname = dir_format % (self.mod_name,
                                      str(today.year),
                                      str(today.month),
                                      str(today.day),
                                      str(today.hour),
                                      str(today.minute),
                                      str(today.second))
                self.directory = os.path.join(self.directory, fname)

                if not os.path.exists(self.directory):
                    os.makedirs(self.directory)
                else:
                    self.clean_directory()

    def getname(self):
        """ get the name of the module """
        return self.mod_name

    def clean_directory(self):
        """ Clean the tmp directory """
        for fname in os.listdir(self.directory):
            fpath = os.path.join(self.directory, fname)
            if os.path.isfile(fpath):
                os.remove(fpath)
            else:
                shutil.rmtree(fpath)

    def send_report(self, subject=None):
        """
        Email the report

        :param subject: subject to set for the email.
                        Default: **ERROR/WARNING for modname**
        :return: None
        """
        if self.smtp_host and self.email:
            if not subject:
                subject = "** ERROR/WARNING for {} **".format(self.mod_name)

            utilities.send_email_netrc(
                self.smtp_host, self.email, subject, self.text_report)


class ScanModule(Module):
    """ Module running on a scan """
    def __init__(self, mod_name, directory, email, text_report):
        """
        Entry point of the derived Module Class for Scan level.

        :param mod_name: Name of the module
        :param directory: Temp directory to store data
        :param email: email address to send report
        :param text_report: string to write for report email
        :return: None
        """
        super(ScanModule, self).__init__(mod_name, directory, email,
                                         text_report)

    def run(self):
        """
        Method to run on one scan
        Implemented in classes.

        :return: None
        """
        raise NotImplementedError()

    def log_warning_error(self, message, scan_info, error=False):
        """
        Print warning or error for a project/subject/session/scan

        :param message: message to print
        :param scan_info: dictionary e.g. <XnatUtils.CachedImageScan>.info()
        :param error: True if the message is an error and not a warning
        :return: None
        """
        error_format = 'ERROR: %s for %s/%s/%s/%s'
        warn_format = 'WARNING: %s for %s/%s/%s/%s'
        if error:
            self.report(error_format % (message,
                                        scan_info['project_id'],
                                        scan_info['subject_label'],
                                        scan_info['session_label'],
                                        scan_info['scan_id']))
        else:
            self.report(warn_format % (message,
                                       scan_info['project_id'],
                                       scan_info['subject_label'],
                                       scan_info['session_label'],
                                       scan_info['scan_id']))


class SessionModule(Module):
    """ Module running on a session """
    def __init__(self, mod_name, directory, email, text_report):
        """
        Entry point of the derived Module Class for Session level.

        :param mod_name: Name of the module
        :param directory: Temp directory to store data
        :param email: email address to send report
        :param text_report: string to write for the report email
        :return: None
        """
        super(SessionModule, self).__init__(mod_name, directory, email,
                                            text_report)

    def run(self):
        """
        Method to run on one session.
        Implemented in classes.

        :return: None
        """
        raise NotImplementedError()

    @staticmethod
    def has_flag_resource(csess, flag_resource):
        """
        Check if the session has the flag_resource

        :param csess: CachedImageSession object (see Xnatutils)
        :param flag_resource: resource to verify its existence on XNAT
        :return: True if the resource exists, False otherwise.
        """
        sess_res_list = csess.get_resources()
        flagres_list = [res for res in sess_res_list
                        if res['label'] == flag_resource]
        if len(flagres_list) > 0:
            LOGGER.debug('Already run')
            return False

        return True

    def log_warning_error(self, message, sess_info, error=False):
        """
        Print warning or error for a project/subject/session

        :param message: message to print
        :param sess_info: dictionary e.g. <XnatUtils.CachedImageSession>.info()
        :param error: True if the message is an error and not a warning
        :return: None
        """
        error_format = 'ERROR: %s for %s/%s/%s'
        warn_format = 'WARNING: %s for %s/%s/%s'
        if error:
            self.report(error_format.format(message,
                                            sess_info['project_id'],
                                            sess_info['subject_label'],
                                            sess_info['session_label']))
        else:
            self.report(warn_format.format(message,
                                           sess_info['project_id'],
                                           sess_info['subject_label'],
                                           sess_info['session_label']))


def modules_by_type(mod_list):
    """
    Method to separate scan modules from session modules

    :param mod_list: list of scan/session modules classes.
    :return: list of session modules, list of scan modules
    """
    sess_mod_list = list()
    scan_mod_list = list()

    # Build list of processors by type
    if mod_list is not None:
        for mod in mod_list:
            if issubclass(mod.__class__, ScanModule):
                scan_mod_list.append(mod)
            elif issubclass(mod.__class__, SessionModule):
                sess_mod_list.append(mod)
            else:
                LOGGER.warn('unknown module type: %s' % mod)

    return sess_mod_list, scan_mod_list



import socket
import os


def check_lockfile(file, logger):
    # Try to read host-PID from lockfile
    try:
        with open(file, 'r') as f:
            line = f.readline()

        host, pid = line.split('-')
        pid = int(pid)

        # Compare host to current host
        this_host = socket.gethostname().split('.')[0]
        if host != this_host:
            logger.debug('different host, cannot check PID:{}', format(file))
        elif pid_exists(pid):
            logger.debug('host matches and PID exists:{}'.format(str(pid)))
        else:
            logger.debug('host matches and PID not running, deleting lockfile')
            os.remove(file)
    except IOError:
        logger.debug('failed to read from lock file:{}'.format(file))
    except ValueError:
        logger.debug('failed to parse lock file:{}'.format(file))


def clean_lockfiles(lock_dir, logger):
    lock_list = os.listdir(lock_dir)

    # Make full paths
    lock_list = [os.path.join(lock_dir, f) for f in lock_list]

    # Check each lock file
    for file in lock_list:
        logger.debug('checking lock file:{}'.format(file))
        check_lockfile(file, logger)


def pid_exists(pid):
    if pid < 0:
        return False   # NOTE: pid == 0 returns True
    try:
        os.kill(pid, 0)
    except ProcessLookupError:   # errno.ESRCH
        return False  # No such process
    except PermissionError:  # errno.EPERM
        return True  # Operation not permitted (i.e., process exists)
    else:
        return True  # no error, we can send a signal to the process


def lock_flagfile(lock_file):
    """
    Create the flagfile to lock the process

    :param lock_file: flag file use to lock the process
    :return: True if the file didn't exist, False otherwise
    """
    if os.path.exists(lock_file):
        return False
    else:
        open(lock_file, 'w').close()

        # Write hostname-PID to lock file
        _pid = os.getpid()
        _host = socket.gethostname().split('.')[0]
        with open(lock_file, 'w') as f:
            f.write('{}-{}'.format(_host, _pid))

        return True


def unlock_flagfile(lock_file):
    """
    Remove the flagfile to unlock the process

    :param lock_file: flag file use to lock the process
    :return: None
    """
    if os.path.exists(lock_file):
        os.remove(lock_file)



import re


SGP_PATTERN = '^\w+-x-\w+-x-\w+_v[0-9]+-x-[0-9a-f]+$'


def full_label(project, subject, session, assessor):
    return '-x-'.join([project, subject, session, assessor])


def full_label_from_assessor(assessor):
    assessor_label = assessor.label()
    if '-x-' in assessor_label:
        return assessor_label

    components = list()
    components.append(assessor_label)
    entity = assessor
    while True:
        entity = entity.parent()
        if entity is None:
            break
        components.append(entity.label())
    return '-x-'.join(reversed(components))


def parse_full_assessor_name(assessor_name):
    elements = assessor_name.split('-x-')
    assrdict = dict()

    if is_sgp_assessor(assessor_name):
        assrdict = dict(list(zip([
            'project_id', 'subject_label', 'session_label', 'label'],
            [elements[0], elements[1], '', assessor_name])))
    elif len(elements) == 5:
        # relabel is in use or old style with scan id in label
        assrdict = dict(list(zip(
            ['project_id', 'subject_label', 'session_label', 'label'],
            [elements[0], elements[1], elements[2], assessor_name])))
    elif len(elements) == 4:
        if len(elements[3]) == 36:
            # new style label with uuid
            assrdict = dict(list(zip(
                ['project_id', 'subject_label', 'session_label', 'label'],
                elements)))
        else:
            # old style label
            assrdict = dict(list(zip(
                ['project_id', 'subject_label', 'session_label', 'label'],
                [elements[0], elements[1], elements[2], assessor_name])))
    else:
        raise ValueError(("'assessor_name' parameter '{}' is not a valid full "
                          "assessor name".format(assessor_name)))

    return assrdict


def is_sgp_assessor(assessor):
    # Try to match the assessor label with the SGP pattern
    return re.match(SGP_PATTERN, assessor)



""" rcq implements a job queue for DAX in REDCap. Launch/Update/Finish, no build or upload here."""
import logging

from ..utilities import get_this_instance
from ..XnatUtils import get_interface

from .taskbuilder import TaskBuilder
from .tasklauncher import TaskLauncher
from .taskqueue import TaskQueue
from .analysislauncher import AnalysisLauncher


logger = logging.getLogger('manager.rcq')


# LAUNCH: requires TaskQueue from projects_redcap and slurm
# BUILD: requires Processors and TaskQueue from projects_redcap and xnat
# UPDATE: required projects_redcap and slurm
# SYNC: requires TaskQueue from projects_redcap and xnat


# TaskBuilder:
# instances_redcap
# projects_redcap
# xnat
# build()


# TaskQueue:
# projects_redcap
# get_open_tasks()
# add_task()
# apply_updates()


def update(rc, instances, build_enabled=True, launch_enabled=True, projects=None):
    logger.info('connecting to redcap')
    def_field = rc.def_field

    if projects is None:
        projects = [x[def_field] for x in rc.export_records(fields=[def_field])]

    instance_settings = _load_instance_settings(instances)
    if not instance_settings:
        logger.debug('no settings found for this instance')
        return

    logger.debug(f'instance_settings={instance_settings}')

    yamldir = instance_settings['main_processorlib']
    logger.debug(f'yamldir={yamldir}')

    with get_interface() as xnat:

        logger.info('Running update of analyses')
        try:
            AnalysisLauncher(xnat, rc, instance_settings).update(
                projects,
                launch_enabled=launch_enabled
            )
        except Exception as err:
            logger.error(f'analyses update failed:{err}')
            pass

        logger.info('Running sync of queue status from XNAT to REDCap')
        TaskQueue(rc).sync(xnat, projects)

        logger.info('Running update of queue from REDCap to SLURM')
        TaskLauncher(rc, instance_settings).update(
                launch_enabled=launch_enabled,
                projects=projects)

        if build_enabled:
            logger.info('Running build of tasks in XNAT and REDCap queue')
            task_builder = TaskBuilder(rc, xnat, yamldir)
            for p in projects:
                logger.debug(f'building:{p}')
                task_builder.update(p)


def _load_instance_settings(instance_redcap):
    """Load DAX settings for current instance from REDCap"""
    instance_name = get_this_instance()
    logger.debug(f'instance={instance_name}')

    # Return the record associated with this instance_name
    records = instance_redcap.export_records(records=[instance_name], raw_or_label='label')
    if len(records) == 0:
        settings = {}
    else:
        settings = records[0]

    return settings





def generate_yaml(procname="Proc",
                  procversion="1_0_0",
                  scans=[],
                  assessors=[]):

    processor_text_ = ('---\n'
                       'inputs:\n'
                       '  default:\n'
                       '    spider_path: ./spiderpath/Spider_{procname}_v{procversion}.py\n'
                       '    working_dir: ./workingdir\n'
                       '    nypype_exe: not_a_real_script.py\n'
                       '    db: not_a_real_db_path/no.db\n'
                       '  xnat:\n'
                       '{scans_block}'
                       '{assessors_block}'
                       '{command}\n'
                       'attrs:\n'
                       '  suffix:\n'
                       '  xsitype: proc:genProcData\n'
                       '  walltime: 24:00:00\n'
                       '  memory: 3850\n'
                       '  ppn: 4\n'
                       '  env: /envpath/not_a_real_env_path.sh\n'
                       '  type: {scan_type}\n'
                       '  scan_nb: {scan_nb}\n')

    cmd_text_ = \
        'command: python {spider_path} --t1 {t1} --dbt {db} --exe {nipype_exe}'

    scans_text_ = '    scans:\n{}'
    scan_text_ = ('      - name: {name}\n'
                  '        types: {types}\n'
                  '{qc}'
                  '{select}'
                  '{select_session}'
                  '{res_block}')

    asrs_text_ = '    assessors:\n{}'
    asr_text_ = ('      - name: {name}\n'
                 '        proctypes: {types}\n'
                 '{qc}'
                 '{select}'
                 '{select_session}'
                 '{res_block}')

    res_block_ = '        resources:\n{}'
    res_text_ = ('          - resource: {res_name}\n'
                 '            varname: {var_name}\n'
                 '{required}')

    req_text_ = '            required: {}\n'

    qc_text_ = '        needs_qc: {}\n'

    select_text_ = '        select: {}\n'
    select_session_text_ = '        select-session: {}\n'

    proc_a_cmd_ =\
        'command: python {spider_path} --t1 {t1} --dbt {db} --exe {nipype_exe}'

    proc_a_scan_text = ('    scans:\n'
                        '      - name: scan1\n'
                        '        types: T1\n'
                        '{qc}'
                        '        resources:\n'
                        '          - resource: NIFTI\n'
                        '            varname: t1\n')

    def generate_input_block(artefact_type, inputs):

        if not artefact_type in ['scan', 'assessor']:
            raise RuntimeError(
                'artefact_type must be one of ''scan'' or ''assessor''')

        if artefact_type == 'scan':
            input_block_ = scans_text_
            input_text_ = scan_text_
        else:
            input_block_ = asrs_text_
            input_text_ = asr_text_

        # create the input block
        if len(inputs) == 0:
            input_block = ''
        else:
            input_entries = []
            for i in inputs:

                # create the select keyword text
                select_value = i.get('select', None)
                if select_value != None:
                    select = select_text_.format(select_value)
                else:
                    select = ''

                # create the select-session keyword text
                select_session_value = i.get('select-session', None)
                if select_session_value != None:
                    select_session =\
                        select_session_text_.format(select_session_value)
                else:
                    select_session = ''

                # create the resource block
                if len(i['resources']) == 0:
                    res_block = ''
                else:
                    res_entries = []
                    for r in i['resources']:
                        required_value = r.get('required', None)
                        if required_value == True:
                            req_text = req_text_.format('True')
                        elif required_value == False:
                            req_text = req_text_.format('False')
                        else:
                            req_text = ''
                        res_entries.append(res_text_.format(res_name=r['type'],
                                                            var_name=r['name'],
                                                            required=req_text))
                    res_block =\
                        res_block_.format(''.join(res_entries))

                qc_value = i.get('qc', None)
                if qc_value == True:
                    qc_text = qc_text_.format('True')
                elif qc_value == False:
                    qc_text = qc_text_.format('False')
                else:
                    qc_text = ''

                if artefact_type == 'scan':
                    input_entries.append(
                        input_text_.format(name=i['name'],
                                           types=i['types'],
                                           select=select,
                                           select_session=select_session,
                                           qc=qc_text,
                                           res_block=res_block))
                else:
                    input_entries.append(
                        input_text_.format(name=i['name'],
                                           types=i['types'],
                                           select=select,
                                           select_session=select_session,
                                           qc=qc_text,
                                           res_block=res_block))

            input_block = input_block_.format(''.join(input_entries))

        return input_block

    if not isinstance(scans, list):
        raise ValueError('scans must be a list')
    if not isinstance(assessors, list):
        raise ValueError('assessors must be a list')

    scan_block = generate_input_block('scan', scans)
    asr_block = generate_input_block('assessor', assessors)

    processor = processor_text_.format(procname=procname,
                                       procversion=procversion,
                                       scans_block=scan_block,
                                       assessors_block=asr_block,
                                       command=cmd_text_,
                                       scan_type='scan',
                                       scan_nb='scan1')
    return processor



proc_a = generate_yaml(
    scans=[
        {
            'name': 'scan1', 'types': 'T1', 'select': 'foreach', 'qc': None,
            'resources': [
                {'type': 'NIFTI', 'name': 't1', 'required': None}]
        },
        {
            'name': 'scan2', 'types': 'FLAIR', 'select': None, 'qc': None,
            'resources': [
                {'type': 'NIFTI', 'name': 'flair', 'required': None}]
        }
    ],
    assessors=[
        {
            'name': 'asr1', 'types': 'proc1', 'select': 'foreach', 'qc': None,
            'resources': [
                {'type': 'SEG', 'name': 'seg', 'required': None},
                {'type': 'STUFF', 'name': 'stuff', 'required': None}]
        }
    ]
)



scan_gif_parcellation_yaml = """
---
inputs:
  default:
    spider_path: /home/dax/Xnat-management/comic100_dax_config/pipelines/GIF_parcellation/v3.0.0/Spider_GIF_Parcellation_v3_0_0.py
    working_dir: /scratch0/dax/
    nipype_exe: perform_gif_propagation.py
    db: /share/apps/cmic/GIF/db/db.xml
  xnat:
    scans:
      - name: scan1
        types: T1w,MPRAGE,T1,T1W
        resources:
          - resource: NIFTI
            varname: t1
command: python {spider_path} --t1 {t1} --dbt {db} --exe {nipype_exe}
attrs:
  suffix:
  xsitype: proc:genProcData
  walltime: 24:00:00
  memory: 3850
  ppn: 4
  env: /share/apps/cmic/NiftyPipe/v2.0/setup_v2.0.sh
  type: scan
  scan_nb: scan1
"""


scan_brain_tiv_from_gif_yaml = """
---
inputs:
  default:
    spider_path: /home/dax/Xnat-management/comic100_dax_config/pipelines/BrainTivFromGIF/v1.0.0/Spider_BrainTivFromGIF_v1_0_0.py
    working_dir: /scratch0/dax/
    nipype_exe: perform_brain_tiv_from_gif.py
  xnat:
    scans:
      - name: scan1
        types: T1W,T1w,MPRAGE 
        resources:
          - resource: NIFTI
            varname: T1
    assessors:
      - name: assessor1
        proctypes: GIF_Parcellation_v3
        resources:
          - resource: SEG
            varname: seg
            required: True
command: python {spider_path} --exe {nipype_exe} --seg {seg}
attrs:
  suffix:
  xsitype: proc:genProcData
  walltime: 01:00:00
  memory: 4096
  ppn: 1
  env: /share/apps/cmic/NiftyPipe/v2.0/setup_v2.0.sh
  type: scan
  scan_nb: scan1 
"""
